"""Benchmark attrs containing other attrs classes."""
import attr
import pytest

from cattr import BaseConverter, Converter, UnstructureStrategy


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
@pytest.mark.parametrize(
    "unstructure_strat", [UnstructureStrategy.AS_DICT, UnstructureStrategy.AS_TUPLE]
)
def test_unstructure_attrs_nested(benchmark, converter_cls, unstructure_strat):
    c = converter_cls(unstruct_strat=unstructure_strat)

    @attr.define
    class InnerA:
        a: int
        b: float
        c: str
        d: bytes

    @attr.define
    class InnerB:
        a: int
        b: float
        c: str
        d: bytes

    @attr.define
    class InnerC:
        a: int
        b: float
        c: str
        d: bytes

    @attr.define
    class InnerD:
        a: int
        b: float
        c: str
        d: bytes

    @attr.define
    class InnerE:
        a: int
        b: float
        c: str
        d: bytes

    @attr.define
    class Outer:
        a: InnerA
        b: InnerB
        c: InnerC
        d: InnerD
        e: InnerE

    inst = Outer(
        InnerA(1, 1.0, "one", "one".encode()),
        InnerB(2, 2.0, "two", "two".encode()),
        InnerC(3, 3.0, "three", "three".encode()),
        InnerD(4, 4.0, "four", "four".encode()),
        InnerE(5, 5.0, "five", "five".encode()),
    )

    benchmark(c.unstructure, inst)


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
@pytest.mark.parametrize(
    "unstructure_strat", [UnstructureStrategy.AS_DICT, UnstructureStrategy.AS_TUPLE]
)
def test_unstruct_attrs_deep_nest(benchmark, converter_cls, unstructure_strat):
    c = converter_cls(unstruct_strat=unstructure_strat)

    @attr.define
    class InnerA:
        a: int
        b: float
        c: str
        d: bytes

    @attr.define
    class InnerB:
        a: InnerA
        b: InnerA
        c: InnerA
        d: InnerA

    @attr.define
    class InnerC:
        a: InnerB
        b: InnerB
        c: InnerB
        d: InnerB

    @attr.define
    class InnerD:
        a: InnerC
        b: InnerC
        c: InnerC
        d: InnerC

    @attr.define
    class InnerE:
        a: InnerD
        b: InnerD
        c: InnerD
        d: InnerD

    @attr.define
    class Outer:
        a: InnerE
        b: InnerE
        c: InnerE
        d: InnerE

    make_inner_a = lambda: InnerA(1, 1.0, "one", "one".encode())
    make_inner_b = lambda: InnerB(*[make_inner_a() for _ in range(4)])
    make_inner_c = lambda: InnerC(*[make_inner_b() for _ in range(4)])
    make_inner_d = lambda: InnerD(*[make_inner_c() for _ in range(4)])
    make_inner_e = lambda: InnerE(*[make_inner_d() for _ in range(4)])

    inst = Outer(*[make_inner_e() for _ in range(4)])

    benchmark(c.unstructure, inst)



import pytest

from cattr import BaseConverter, Converter


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
def test_unstructure_int(benchmark, converter_cls):
    c = converter_cls()

    benchmark(c.unstructure, 5)


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
def test_unstructure_float(benchmark, converter_cls):
    c = converter_cls()

    benchmark(c.unstructure, 15.0)



"""Utilities for union (sum type) disambiguation."""
from collections import OrderedDict, defaultdict
from functools import reduce
from operator import or_
from typing import Any, Callable, Dict, Mapping, Optional, Set, Type, Union

from attrs import NOTHING, fields, fields_dict

from ._compat import get_args, get_origin, has, is_literal, is_union_type

__all__ = ("is_supported_union", "create_default_dis_func")

NoneType = type(None)


def is_supported_union(typ: Type) -> bool:
    """Whether the type is a union of attrs classes."""
    return is_union_type(typ) and all(
        e is NoneType or has(get_origin(e) or e) for e in typ.__args__
    )


def create_default_dis_func(
    *classes: Type[Any], use_literals: bool = True
) -> Callable[[Mapping[Any, Any]], Optional[Type[Any]]]:
    """Given attrs classes, generate a disambiguation function.

    The function is based on unique fields or unique values.

    :param use_literals: Whether to try using fields annotated as literals for
        disambiguation.
    """
    if len(classes) < 2:
        raise ValueError("At least two classes required.")

    # first, attempt for unique values
    if use_literals:
        # requirements for a discriminator field:
        # (... TODO: a single fallback is OK)
        #  - it must always be enumerated
        cls_candidates = [
            {at.name for at in fields(get_origin(cl) or cl) if is_literal(at.type)}
            for cl in classes
        ]

        # literal field names common to all members
        discriminators: Set[str] = cls_candidates[0]
        for possible_discriminators in cls_candidates:
            discriminators &= possible_discriminators

        best_result = None
        best_discriminator = None
        for discriminator in discriminators:
            # maps Literal values (strings, ints...) to classes
            mapping = defaultdict(list)

            for cl in classes:
                for key in get_args(
                    fields_dict(get_origin(cl) or cl)[discriminator].type
                ):
                    mapping[key].append(cl)

            if best_result is None or max(len(v) for v in mapping.values()) <= max(
                len(v) for v in best_result.values()
            ):
                best_result = mapping
                best_discriminator = discriminator

        if (
            best_result
            and best_discriminator
            and max(len(v) for v in best_result.values()) != len(classes)
        ):
            final_mapping = {
                k: v[0] if len(v) == 1 else Union[tuple(v)]
                for k, v in best_result.items()
            }

            def dis_func(data: Mapping[Any, Any]) -> Optional[Type]:
                if not isinstance(data, Mapping):
                    raise ValueError("Only input mappings are supported.")
                return final_mapping[data[best_discriminator]]

            return dis_func

    # next, attempt for unique keys

    # NOTE: This could just as well work with just field availability and not
    #  uniqueness, returning Unions ... it doesn't do that right now.
    cls_and_attrs = [
        (cl, {at.name for at in fields(get_origin(cl) or cl)}) for cl in classes
    ]
    if len([attrs for _, attrs in cls_and_attrs if len(attrs) == 0]) > 1:
        raise ValueError("At least two classes have no attributes.")
    # TODO: Deal with a single class having no required attrs.
    # For each class, attempt to generate a single unique required field.
    uniq_attrs_dict: Dict[str, Type] = OrderedDict()
    cls_and_attrs.sort(key=lambda c_a: -len(c_a[1]))

    fallback = None  # If none match, try this.

    for i, (cl, cl_reqs) in enumerate(cls_and_attrs):
        other_classes = cls_and_attrs[i + 1 :]
        if other_classes:
            other_reqs = reduce(or_, (c_a[1] for c_a in other_classes))
            uniq = cl_reqs - other_reqs
            if not uniq:
                m = f"{cl} has no usable unique attributes."
                raise ValueError(m)
            # We need a unique attribute with no default.
            cl_fields = fields(get_origin(cl) or cl)
            for attr_name in uniq:
                if getattr(cl_fields, attr_name).default is NOTHING:
                    break
            else:
                raise ValueError(f"{cl} has no usable non-default attributes.")
            uniq_attrs_dict[attr_name] = cl
        else:
            fallback = cl

    def dis_func(data: Mapping[Any, Any]) -> Optional[Type]:
        if not isinstance(data, Mapping):
            raise ValueError("Only input mappings are supported.")
        for k, v in uniq_attrs_dict.items():
            if k in data:
                return v
        return fallback

    return dis_func


create_uniq_field_dis_func = create_default_dis_func



"""Useful internal functions."""
from typing import NoReturn, Type, TypeVar

from .errors import StructureHandlerNotFoundError

T = TypeVar("T")


def identity(obj: T) -> T:
    """The identity function."""
    return obj


def raise_error(_, cl: Type) -> NoReturn:
    """At the bottom of the condition stack, we explode if we can't handle it."""
    msg = f"Unsupported type: {cl!r}. Register a structure hook for it."
    raise StructureHandlerNotFoundError(msg, type_=cl)



import itertools
from typing import Union

import pytest
from attrs import define
from hypothesis import given
from hypothesis.strategies import integers

from cattrs import BaseConverter
from cattrs.strategies import use_class_methods


@define
class Base:
    a: int


class Structure(Base):
    @classmethod
    def _structure(cls, data: dict):
        return cls(data["b"])  # expecting "b", not "a"


class Unstructure(Base):
    def _unstructure(self):
        return {"c": self.a}  # unstructuring as "c", not "a"


class Both(Structure, Unstructure):
    pass


@pytest.fixture
def get_converter(converter: BaseConverter):
    def aux(structure: str, unstructure: str) -> BaseConverter:
        use_class_methods(converter, structure, unstructure)
        return converter

    return aux


@pytest.mark.parametrize(
    "cls,structure_method,unstructure_method",
    itertools.product(
        [Structure, Unstructure, Both],
        ["_structure", "_undefined", None],
        ["_unstructure", "_undefined", None],
    ),
)
def test_not_nested(get_converter, structure_method, unstructure_method, cls) -> None:
    converter = get_converter(structure_method, unstructure_method)

    assert converter.structure(
        {
            "b"
            if structure_method == "_structure" and hasattr(cls, "_structure")
            else "a": 42
        },
        cls,
    ) == cls(42)

    assert converter.unstructure(cls(42)) == {
        "c"
        if unstructure_method == "_unstructure" and hasattr(cls, "_unstructure")
        else "a": 42
    }


@given(integers(1, 5))
def test_nested_roundtrip(depth):
    @define
    class Nested:
        a: Union["Nested", None]
        c: int

        @classmethod
        def _structure(cls, data, conv):
            b = data["b"]
            return cls(None if b is None else conv.structure(b, cls), data["c"])

        def _unstructure(self, conv):
            return {"b": conv.unstructure(self.a), "c": self.c}

        @staticmethod
        def create(depth: int) -> Union["Nested", None]:
            return None if depth == 0 else Nested(Nested.create(depth - 1), 42)

    structured = Nested.create(depth)

    converter = BaseConverter()
    use_class_methods(converter, "_structure", "_unstructure")
    assert structured == converter.structure(converter.unstructure(structured), Nested)



from enum import IntEnum
from typing import Dict, List, Mapping, MutableMapping

import attr
import pytest

from cattr import BaseConverter, Converter, UnstructureStrategy


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
@pytest.mark.parametrize(
    "unstructure_strat", [UnstructureStrategy.AS_DICT, UnstructureStrategy.AS_TUPLE]
)
def test_unstructure_attrs_lists(benchmark, converter_cls, unstructure_strat):
    """
    Benchmark a large (30 attributes) attrs class containing lists of
    primitives.
    """

    class E(IntEnum):
        ONE = 1
        TWO = 2

    @attr.define
    class C:
        a: List[int]
        b: List[float]
        c: List[str]
        d: List[bytes]
        e: List[E]
        f: List[int]
        g: List[float]
        h: List[str]
        i: List[bytes]
        j: List[E]
        k: List[int]
        l: List[float]
        m: List[str]
        n: List[bytes]
        o: List[E]
        p: List[int]
        q: List[float]
        r: List[str]
        s: List[bytes]
        t: List[E]
        u: List[int]
        v: List[float]
        w: List[str]
        x: List[bytes]
        y: List[E]
        z: List[int]
        aa: List[float]
        ab: List[str]
        ac: List[bytes]
        ad: List[E]

    c = converter_cls(unstruct_strat=unstructure_strat)

    benchmark(
        c.unstructure,
        C(
            [1] * 3,
            [1.0] * 3,
            ["a small string"] * 3,
            ["test".encode()] * 3,
            [E.ONE] * 3,
            [2] * 3,
            [2.0] * 3,
            ["a small string"] * 3,
            ["test".encode()] * 3,
            [E.TWO] * 3,
            [3] * 3,
            [3.0] * 3,
            ["a small string"] * 3,
            ["test".encode()] * 3,
            [E.ONE] * 3,
            [4] * 3,
            [4.0] * 3,
            ["a small string"] * 3,
            ["test".encode()] * 3,
            [E.TWO] * 3,
            [5] * 3,
            [5.0] * 3,
            ["a small string"] * 3,
            ["test".encode()] * 3,
            [E.ONE] * 3,
            [6] * 3,
            [6.0] * 3,
            ["a small string"] * 3,
            ["test".encode()] * 3,
            [E.TWO] * 3,
        ),
    )


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
@pytest.mark.parametrize(
    "unstructure_strat", [UnstructureStrategy.AS_DICT, UnstructureStrategy.AS_TUPLE]
)
def test_unstructure_attrs_mappings(benchmark, converter_cls, unstructure_strat):
    """
    Benchmark an attrs class containing mappings.
    """

    @attr.frozen
    class FrozenCls:
        a: int

    @attr.define
    class C:
        a: Mapping[int, str]
        b: Dict[float, bytes]
        c: MutableMapping[int, FrozenCls]

    c = converter_cls(unstruct_strat=unstructure_strat)

    benchmark(
        c.unstructure,
        C(
            {i: str(i) for i in range(30)},
            {float(i): bytes(i) for i in range(30)},
            {i: FrozenCls(i) for i in range(30)},
        ),
    )


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
def test_structure_attrs_mappings(benchmark, converter_cls):
    """
    Benchmark an attrs class containing mappings.
    """

    @attr.frozen
    class FrozenCls:
        a: int

    @attr.define
    class C:
        a: Mapping[int, str]
        b: Dict[float, bytes]
        c: MutableMapping[int, FrozenCls]

    c = converter_cls()

    inst = C(
        {i: str(i) for i in range(30)},
        {float(i): bytes(i) for i in range(30)},
        {i: FrozenCls(i) for i in range(30)},
    )
    raw = c.unstructure(inst)

    benchmark(c.structure, raw, C)



"""Tests for auto-disambiguators."""
from typing import Literal, Union

import pytest
from attrs import NOTHING, asdict, define, field, fields
from hypothesis import HealthCheck, assume, given, settings

from cattrs import Converter
from cattrs.disambiguators import (
    create_default_dis_func,
    create_uniq_field_dis_func,
    is_supported_union,
)

from .untyped import simple_classes


def test_edge_errors():
    """Edge input cases cause errors."""

    @define
    class A:
        pass

    with pytest.raises(ValueError):
        # Can't generate for only one class.
        create_uniq_field_dis_func(A)

    with pytest.raises(ValueError):
        create_default_dis_func(A)

    @define
    class B:
        pass

    with pytest.raises(ValueError):
        # No fields on either class.
        create_uniq_field_dis_func(A, B)

    with pytest.raises(ValueError):
        create_default_dis_func(A, B)

    @define
    class C:
        a = field()

    @define
    class D:
        a = field()

    with pytest.raises(ValueError):
        # No unique fields on either class.
        create_uniq_field_dis_func(C, D)

    with pytest.raises(ValueError):
        # No discriminator candidates
        create_default_dis_func(C, D)

    @define
    class E:
        pass

    @define
    class F:
        b = None

    with pytest.raises(ValueError):
        # no usable non-default attributes
        create_uniq_field_dis_func(E, F)

    @define
    class G:
        x: Literal[1]

    @define
    class H:
        x: Literal[1]

    with pytest.raises(ValueError):
        # The discriminator chosen does not actually help
        create_default_dis_func(C, D)


@given(simple_classes(defaults=False))
def test_fallback(cl_and_vals):
    """The fallback case works."""
    cl, vals, kwargs = cl_and_vals

    assume(fields(cl))  # At least one field.

    @define
    class A:
        pass

    fn = create_uniq_field_dis_func(A, cl)

    assert fn({}) is A
    assert fn(asdict(cl(*vals, **kwargs))) is cl

    attr_names = {a.name for a in fields(cl)}

    if "xyz" not in attr_names:
        assert fn({"xyz": 1}) is A  # Uses the fallback.


@settings(suppress_health_check=[HealthCheck.filter_too_much, HealthCheck.too_slow])
@given(simple_classes(), simple_classes())
def test_disambiguation(cl_and_vals_a, cl_and_vals_b):
    """Disambiguation should work when there are unique required fields."""
    cl_a, vals_a, kwargs_a = cl_and_vals_a
    cl_b, vals_b, kwargs_b = cl_and_vals_b

    req_a = {a.name for a in fields(cl_a)}
    req_b = {a.name for a in fields(cl_b)}

    assume(len(req_a))
    assume(len(req_b))

    assume((req_a - req_b) or (req_b - req_a))
    for attr_name in req_a - req_b:
        assume(getattr(fields(cl_a), attr_name).default is NOTHING)
    for attr_name in req_b - req_a:
        assume(getattr(fields(cl_b), attr_name).default is NOTHING)

    fn = create_uniq_field_dis_func(cl_a, cl_b)

    assert fn(asdict(cl_a(*vals_a, **kwargs_a))) is cl_a


# not too sure of properties of `create_default_dis_func`
def test_disambiguate_from_discriminated_enum():
    # can it find any discriminator?
    @define
    class A:
        a: Literal[0]

    @define
    class B:
        a: Literal[1]

    fn = create_default_dis_func(A, B)
    assert fn({"a": 0}) is A
    assert fn({"a": 1}) is B

    # can it find the better discriminator?
    @define
    class C:
        a: Literal[0]
        b: Literal[1]

    @define
    class D:
        a: Literal[0]
        b: Literal[0]

    fn = create_default_dis_func(C, D)
    assert fn({"a": 0, "b": 1}) is C
    assert fn({"a": 0, "b": 0}) is D

    # can it handle multiple tiers of discriminators?
    # (example inspired by Discord's gateway's discriminated union)
    @define
    class E:
        op: Literal[1]

    @define
    class F:
        op: Literal[0]
        t: Literal["MESSAGE_CREATE"]

    @define
    class G:
        op: Literal[0]
        t: Literal["MESSAGE_UPDATE"]

    fn = create_default_dis_func(E, F, G)
    assert fn({"op": 1}) is E
    assert fn({"op": 0, "t": "MESSAGE_CREATE"}) is Union[F, G]

    # can it handle multiple literals?
    @define
    class H:
        a: Literal[1]

    @define
    class J:
        a: Literal[0, 1]

    @define
    class K:
        a: Literal[0]

    fn = create_default_dis_func(H, J, K)
    assert fn({"a": 1}) is Union[H, J]
    assert fn({"a": 0}) is Union[J, K]


def test_default_no_literals():
    """The default disambiguator can skip literals."""

    @define
    class A:
        a: Literal["a"] = "a"

    @define
    class B:
        a: Literal["b"] = "b"

    default = create_default_dis_func(A, B)  # Should work.
    assert default({"a": "a"}) is A

    with pytest.raises(ValueError):
        create_default_dis_func(A, B, use_literals=False)

    @define
    class C:
        b: int
        a: Literal["a"] = "a"

    @define
    class D:
        a: Literal["b"] = "b"

    default = create_default_dis_func(C, D)  # Should work.
    assert default({"a": "a"}) is C

    no_lits = create_default_dis_func(C, D, use_literals=False)
    assert no_lits({"a": "a", "b": 1}) is C
    assert no_lits({"a": "a"}) is D


def test_converter_no_literals(converter: Converter):
    """A converter can be configured to skip literals."""
    from functools import partial

    converter.register_structure_hook_factory(
        is_supported_union,
        partial(converter._gen_attrs_union_structure, use_literals=False),
    )

    @define
    class C:
        b: int
        a: Literal["a"] = "a"

    @define
    class D:
        a: Literal["b"] = "b"

    assert converter.structure({}, Union[C, D]) == D()



from enum import IntEnum

import attr
import pytest

from cattr import BaseConverter, Converter, UnstructureStrategy


class E(IntEnum):
    ONE = 1
    TWO = 2


@attr.define
class C:
    a: int
    b: float
    c: str
    d: bytes
    e: E
    f: int
    g: float
    h: str
    i: bytes
    j: E
    k: int
    l: float
    m: str
    n: bytes
    o: E
    p: int
    q: float
    r: str
    s: bytes
    t: E
    u: int
    v: float
    w: str
    x: bytes
    y: E
    z: int
    aa: float
    ab: str
    ac: bytes
    ad: E


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
@pytest.mark.parametrize(
    "unstructure_strat", [UnstructureStrategy.AS_DICT, UnstructureStrategy.AS_TUPLE]
)
def test_unstructure_attrs_primitives(benchmark, converter_cls, unstructure_strat):
    """Benchmark a large (30 attributes) attrs class containing primitives."""

    c = converter_cls(unstruct_strat=unstructure_strat)

    benchmark(
        c.unstructure,
        C(
            1,
            1.0,
            "a small string",
            "test".encode(),
            E.ONE,
            2,
            2.0,
            "a small string",
            "test".encode(),
            E.TWO,
            3,
            3.0,
            "a small string",
            "test".encode(),
            E.ONE,
            4,
            4.0,
            "a small string",
            "test".encode(),
            E.TWO,
            5,
            5.0,
            "a small string",
            "test".encode(),
            E.ONE,
            6,
            6.0,
            "a small string",
            "test".encode(),
            E.TWO,
        ),
    )


@pytest.mark.parametrize("converter_cls", [BaseConverter, Converter])
@pytest.mark.parametrize(
    "unstructure_strat", [UnstructureStrategy.AS_DICT, UnstructureStrategy.AS_TUPLE]
)
def test_structure_attrs_primitives(benchmark, converter_cls, unstructure_strat):
    """Benchmark a large (30 attributes) attrs class containing primitives."""

    c = converter_cls(unstruct_strat=unstructure_strat)

    inst = C(
        1,
        1.0,
        "a small string",
        "test".encode(),
        E.ONE,
        2,
        2.0,
        "a small string",
        "test".encode(),
        E.TWO,
        3,
        3.0,
        "a small string",
        "test".encode(),
        E.ONE,
        4,
        4.0,
        "a small string",
        "test".encode(),
        E.TWO,
        5,
        5.0,
        "a small string",
        "test".encode(),
        E.ONE,
        6,
        6.0,
        "a small string",
        "test".encode(),
        E.TWO,
    )

    raw = c.unstructure(inst)

    benchmark(c.structure, raw, C)



import sys

is_py38 = sys.version_info[:2] == (3, 8)
is_py39_plus = sys.version_info >= (3, 9)
is_py310_plus = sys.version_info >= (3, 10)
is_py311_plus = sys.version_info >= (3, 11)

if is_py38:
    from typing import Dict, List

    List_origin = List
    Dict_origin = Dict


else:
    List_origin = list
    Dict_origin = dict



from typing import Any, NewType, Optional

import pytest
from attrs import define

from cattrs import Converter

from ._compat import is_py310_plus


def test_newtype_optionals(genconverter):
    """Newtype optionals should work."""
    Foo = NewType("Foo", str)

    genconverter.register_unstructure_hook(Foo, lambda v: v.replace("foo", "bar"))

    @define
    class ModelWithFoo:
        total_foo: Foo
        maybe_foo: Optional[Foo]

    assert genconverter.unstructure(ModelWithFoo(Foo("foo"), Foo("is it a foo?"))) == {
        "total_foo": "bar",
        "maybe_foo": "is it a bar?",
    }


@pytest.mark.skipif(not is_py310_plus, reason="3.10+ union syntax")
def test_newtype_modern_optionals(genconverter):
    """Newtype optionals should work."""
    Foo = NewType("Foo", str)

    genconverter.register_unstructure_hook(Foo, lambda v: v.replace("foo", "bar"))

    @define
    class ModelWithFoo:
        total_foo: Foo
        maybe_foo: Foo | None

    assert genconverter.unstructure(ModelWithFoo(Foo("foo"), Foo("is it a foo?"))) == {
        "total_foo": "bar",
        "maybe_foo": "is it a bar?",
    }


def test_optional_any(converter: Converter):
    """Unstructuring Any|None is equivalent to unstructuring as v.__class__."""

    @define
    class A:
        pass

    assert converter.unstructure(A(), Optional[Any]) == {}
